#pragma once

#include <type_traits>

#include <dune/common/std/optional.hh>
#include <dune/functions/common/signature.hh>
#include <dune/functions/gridfunctions/gridviewentityset.hh>

#include <amdis/Operations.hpp>
#include <amdis/gridfunctions/GridFunctionConcepts.hpp>

namespace AMDiS
{
#ifndef DOXYGEN
  template <class Signature, class LocalContext, class Function>
  class AnalyticLocalFunction;
#endif

  template <class R, class D, class LocalContext, class Function>
  class AnalyticLocalFunction<R(D), LocalContext, Function>
  {
  public:
    using Range = R;
    using Domain = D; // LocalDomain

    using Geometry = typename LocalContext::Geometry;

    enum { hasDerivative = true };

  public:
    AnalyticLocalFunction(Function const& fct)
      : fct_{fct}
    {}

    void bind(LocalContext const& element)
    {
      geometry_.emplace(element.geometry());
    }

    void unbind()
    {
      geometry_.reset();
    }

    Range operator()(Domain const& local) const
    {
      assert( bool(geometry_) );
      return fct_(geometry_.value().global(local));
    }

    Function const& fct() const
    {
      return fct_;
    }

  private:
    Function fct_;
    Dune::Std::optional<Geometry> geometry_;
  };

  /// \brief Return the polynomial order of the function f, if a free function
  /// order(f) exists, otherwise return a default value. \relates AnalyticLocalFunction
  template <class Sig, class LocalContext, class F,
    REQUIRES(Concepts::HasOrder<F>)>
  int order(AnalyticLocalFunction<Sig,LocalContext,F> const& lf)
  {
    return order(lf.fct(),1);
  }


  template <class R, class D, class LocalContext, class F>
  auto derivative(AnalyticLocalFunction<R(D),LocalContext,F> const& lf)
  {
    static_assert(Concepts::HasPartial<F>,
      "No partial(_0,...) defined for Functor F of AnalyticLocalFunction.");

    auto df = partial(lf.fct(), index_<0>);

    using RawSignature = typename Dune::Functions::SignatureTraits<R(D)>::RawSignature;
    using DerivativeSignature = typename Dune::Functions::DefaultDerivativeTraits<RawSignature>::Range(D);
    return AnalyticLocalFunction<DerivativeSignature,LocalContext,decltype(df)>{df};
  }


  /// \class AnalyticGridFunction
  /// \brief A Gridfunction that evaluates a function with global coordinates.
  /**
   * \ingroup GridFunctions
   * Implements a GridFunction that wraps a functor around global coordinates.
   *
   * \tparam Function The callable `f=f(x)` with `x in Domain == GlobalCoordinates`
   * \tparam GridView A GridView that models `Dune::GridViewConcept`.
   *
   * **Requirements:**
   * - Function models \ref Concepts::Callable<Domain>
   **/
  template <class Function, class GridView>
  class AnalyticGridFunction
  {
  public:
    using EntitySet = Dune::Functions::GridViewEntitySet<GridView, 0>;
    using Domain = typename EntitySet::GlobalCoordinate;
    using Range = std::decay_t<std::result_of_t<Function(Domain)>>;

    enum { hasDerivative = true };

  private:
    using Element = typename EntitySet::Element;
    using LocalDomain = typename EntitySet::LocalCoordinate;
    using LocalFunction = AnalyticLocalFunction<Range(LocalDomain), Element, Function>;

  public:
    /// \brief Constructor. Stores the function `fct` and creates an `EntitySet`.
    AnalyticGridFunction(Function const& fct, GridView const& gridView)
      : fct_{fct}
      , entitySet_{gridView}
    {}

    /// \brief Return the evaluated functor at global coordinates
    Range operator()(Domain const& x) const
    {
      return fct_(x);
    }

    /// \brief Return the LocalFunction of the AnalyticGridFunction.
    LocalFunction localFunction() const
    {
      return {fct_};
    }

    EntitySet const& entitySet() const
    {
      return entitySet_;
    }

    Function const& fct() const { return fct_; }

  private:
    Function fct_;

    EntitySet entitySet_;
  };


  template <class F, class GV>
  auto localFunction(AnalyticGridFunction<F,GV> const& gf)
  {
    return gf.localFunction();
  }

  /// \brief Return a GridFunction representing the derivative of a functor.
  // [expects: Functor f has free function derivative(f)]
  template <class F, class GV>
  auto derivative(AnalyticGridFunction<F,GV> const& gf)
  {
    static_assert(Concepts::HasPartial<F>,
      "No partial(...,_0) defined for Functor of AnalyticLocalFunction.");

    auto df = partial(gf.fct(), index_<0>);
    return AnalyticGridFunction<decltype(df), GV>{df, gf.entitySet().gridView()};
  }


#ifndef DOXYGEN
  // A pre-GridFunction that just stores the function
  template <class Function>
  struct AnalyticPreGridFunction
  {
    using Self = AnalyticPreGridFunction;

    struct Creator
    {
      template <class GridView>
      static auto create(Self const& self, GridView const& gridView)
      {
        return AnalyticGridFunction<Function, GridView>{self.fct_, gridView};
      }
    };

    Function fct_;
  };

  namespace Traits
  {
    template <class Functor>
    struct IsPreGridFunction<AnalyticPreGridFunction<Functor>>
      : std::true_type {};
  }
#endif


  /// \fn evalAtQP \brief Generator function for AnalyticGridFunction. \relates AnalyticGridfunction
  /**
   * \ingroup GridFunctions
   * Evaluate a functor at global coordinates. See \ref AnalyticGridFunction.
   *
   * **Examples:**
   * - `evalAtQP([](Dune::FieldVector<double, 2> const& x) { return x[0]; })`
   * - `evalAtQP(Operation::TwoNorm{})`
   **/
  template <class Function,
    REQUIRES(Concepts::CallableDomain<Function>)>
  auto evalAtQP(Function const& f)
  {
    return AnalyticPreGridFunction<Function>{f};
  }


  template <class Function>
  struct GridFunctionCreator<Function, std::enable_if_t<Concepts::CallableDomain<Function>>>
  {
    template <class GridView>
    static auto create(Function const& fct, GridView const& gridView)
    {
      return AnalyticGridFunction<Function, GridView>{fct, gridView};
    }
  };

} // end namespace AMDiS
