#pragma once
#include "config.h"
#include <dune/geometry/quadraturerules.hh>

#include <amdis/GridFunctions.hpp>

namespace AMDiS
{
  namespace Impl
  {
    template <class GridFct, class GridView, class QuadProvider>
    auto integrateImpl(GridFct&& gridFct, GridView const& gridView, QuadProvider makeQuad)
    {
      auto localFct = localFunction(gridFct);

      using Range = typename std::decay_t<GridFct>::Range;
      Range result(0);

      for (auto const& element : elements(gridView)) {
        auto geometry = element.geometry();

        localFct.bind(element);
        auto const& quad = makeQuad(element.type(), localFct);

        for (auto const qp : quad)
          result += localFct(qp.position()) * geometry.integrationElement(qp.position()) * qp.weight();
        localFct.unbind();
      }

      return result;
    }

  } // end namespace Impl


  /// Integrate expression with quadrature rule determined by polynomial order of expression
  template <class Expr, class GridView>
  auto integrate(Expr&& expr, GridView const& gridView)
  {
    auto&& gridFct = makeGridFunction(std::forward<Expr>(expr), gridView);
    static const bool expr_has_order = Concepts::HasLocalFunctionOrder<std::decay_t<decltype(gridFct)>>;
    static_assert(expr_has_order,
      "Polynomial degree of expression can not be deduced. You need to provide an explicit value for the quadrature degree or a quadrature rule in `integrate()`.");

    using QuadratureRules = Dune::QuadratureRules<typename GridView::ctype, GridView::dimension>;
    auto makeQuad = [](auto&& type, auto&& localFct) { return QuadratureRules::rule(type, order(localFct)); };

#if AMDIS_HAS_CXX_CONSTEXPR_IF
    if constexpr(expr_has_order)
      return Impl::integrateImpl(gridFct, gridView, makeQuad);
    else
      return 0.0;
#else
    return Dune::Hybrid::ifElse(bool_<expr_has_order>,
      [&](auto) { return Impl::integrateImpl(gridFct, gridView, makeQuad); },
      [ ](auto) { return 0.0; });
#endif
  }

  /// Integrate expression with quadrature rule provided
  template <class Expr, class GridView,
    class QuadratureRule = Dune::QuadratureRule<typename GridView::ctype, GridView::dimension>>
  auto integrate(Expr&& expr, GridView const& gridView, QuadratureRule const& quad)
  {
    auto&& gridFct = makeGridFunction(std::forward<Expr>(expr), gridView);
    return Impl::integrateImpl(std::forward<decltype(gridFct)>(gridFct), gridView,
      [&](auto&&, auto&&) { return quad; });
  }

  /// Integrate expression with quadrature rule determined by provided polynomial `degree`
  template <class Expr, class GridView>
  auto integrate(Expr&& expr, GridView const& gridView, int degree,
                 Dune::QuadratureType::Enum qt = Dune::QuadratureType::GaussLegendre)
  {
    using QuadratureRules = Dune::QuadratureRules<typename GridView::ctype, GridView::dimension>;

    auto&& gridFct = makeGridFunction(std::forward<Expr>(expr), gridView);
    return Impl::integrateImpl(std::forward<decltype(gridFct)>(gridFct), gridView,
      [&](auto const& type, auto&&) { return QuadratureRules::rule(type, degree, qt); });
  }

} // end namespace AMDiS
