#pragma once

#include <amdis/Operations.hpp>
#include <amdis/gridfunctions/FunctorGridFunction.hpp>
#include <amdis/operations/Arithmetic.hpp>
#include <amdis/operations/FieldMatVec.hpp>

namespace AMDiS
{
  /**
    * \addtogroup GridFunctions
    * @{
    **/

  // scalar operations
  // @{

  /// \brief Applies \ref Operation::Negate to GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs>)>
  auto operator-(Lhs&& lhs)
  {
    return invokeAtQP(Operation::Negate{}, std::forward<Lhs>(lhs));
  }

  /// \brief Applies \ref Operation::Plus to GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto operator+(Lhs&& lhs, Rhs&& rhs)
  {
    return invokeAtQP(Operation::Plus{}, std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }

  /// \brief Applies \ref Operation::Minus to GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto operator-(Lhs&& lhs, Rhs&& rhs)
  {
    return invokeAtQP(Operation::Minus{}, std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }

  /// \brief Applies \ref Operation::Multiplies to GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto operator*(Lhs&& lhs, Rhs&& rhs)
  {
    return invokeAtQP(Operation::Multiplies{}, std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }

  /// \brief Applies \ref Operation::Divides to GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto operator/(Lhs&& lhs, Rhs&& rhs)
  {
    return invokeAtQP(Operation::Divides{}, std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }


  /// \brief Applies \ref Operation::Max to GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto max(Lhs&& lhs, Rhs&& rhs)
  {
    return invokeAtQP(Operation::Max{}, std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }

  /// \brief Applies \ref Operation::Min to GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto min(Lhs&& lhs, Rhs&& rhs)
  {
    return invokeAtQP(Operation::Min{}, std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }

  /// \brief Applies \ref Operation::AbsMax to GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto abs_max(Lhs&& lhs, Rhs&& rhs)
  {
    return invokeAtQP(Operation::AbsMax{}, std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }

  /// \brief Applies \ref Operation::AbsMin to GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto abs_min(Lhs&& lhs, Rhs&& rhs)
  {
    return invokeAtQP(Operation::AbsMin{}, std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }

  /// \brief Applies \ref Operation::Sqr to GridFunction. \relates FunctorGridFunction
  template <class T,
    REQUIRES(Concepts::AnyGridFunction<T>)>
  auto sqr(T&& value)
  {
    return invokeAtQP(Operation::Sqr{}, std::forward<T>(value));
  }

  /// \brief Applies \ref Operation::Pow<p> to GridFunction. \relates FunctorGridFunction
  template <int p, class T,
    REQUIRES(Concepts::AnyGridFunction<T>)>
  auto pow(T&& value)
  {
    return invokeAtQP(Operation::Pow<p>{}, std::forward<T>(value));
  }

  /// \brief Applies \ref Operation::Pow_ to GridFunction. \relates FunctorGridFunction
  template <class T,
    REQUIRES(Concepts::AnyGridFunction<T>)>
  auto pow(T&& value, int p)
  {
    return invokeAtQP(Operation::Pow_{p}, std::forward<T>(value));
  }

  /// \brief Applies \ref Operation::Get<I> to GridFunction. \relates FunctorGridFunction
  template <std::size_t I, class T,
    REQUIRES(Concepts::AnyGridFunction<T>)>
  auto get(T&& value)
  {
    return invokeAtQP(Operation::Get<I>{}, std::forward<T>(value));
  }

  /// \brief Applies \ref Operation::Get_ to GridFunction. \relates FunctorGridFunction
  template <class T,
    REQUIRES(Concepts::AnyGridFunction<T>)>
  auto get(T&& value, std::size_t i)
  {
    return invokeAtQP(Operation::Get_{i}, std::forward<T>(value));
  }

  // @}

  // unary vector operations
  // @{

  /// \brief Applies a sum-functor to a vector-valued GridFunction. \relates FunctorGridFunction
  template <class Vec,
    REQUIRES(Concepts::AnyGridFunction<Vec>)>
  auto sum(Vec&& vec)
  {
    return invokeAtQP([](auto const& v) { return sum(v); }, std::forward<Vec>(vec));
  }

  /// \brief Applies \ref Operation::UnaryDot to a vector-valued GridFunction. \relates FunctorGridFunction
  template <class Vec,
    REQUIRES(Concepts::AnyGridFunction<Vec>)>
  auto unary_dot(Vec&& vec)
  {
    return invokeAtQP(Operation::UnaryDot{}, std::forward<Vec>(vec));
  }

  /// \brief Applies a one_norm-functor to a vector-valued GridFunction. \relates FunctorGridFunction
  template <class Vec,
    REQUIRES(Concepts::AnyGridFunction<Vec>)>
  auto one_norm(Vec&& vec)
  {
    return invokeAtQP([](auto const& v) { return one_norm(v); }, std::forward<Vec>(vec));
  }

  /// \brief Applies \ref Operation::TwoNorm to a vector-valued GridFunction. \relates FunctorGridFunction
  template <class Vec,
    REQUIRES(Concepts::AnyGridFunction<Vec>)>
  auto two_norm(Vec&& vec)
  {
    return invokeAtQP(Operation::TwoNorm{}, std::forward<Vec>(vec));
  }

  /// \brief Applies a p_norm-functor to a vector-valued GridFunction. \relates FunctorGridFunction
  template <int p, class Vec,
    REQUIRES(Concepts::AnyGridFunction<Vec>)>
  auto p_norm(Vec&& vec)
  {
    return invokeAtQP([](auto const& v) { return p_norm<p>(v); }, std::forward<Vec>(vec));
  }

  /// \brief Applies a infty_norm-functor to a vector-valued GridFunction. \relates FunctorGridFunction
  template <class Vec,
    REQUIRES(Concepts::AnyGridFunction<Vec>)>
  auto infty_norm(Vec&& vec)
  {
    return invokeAtQP([](auto const& v) { return infty_norm(v); }, std::forward<Vec>(vec));
  }

  /// \brief Applies a trans-functor to a matrix-valued GridFunction.\relates FunctorGridFunction
  template <class Mat,
    REQUIRES(Concepts::AnyGridFunction<Mat>)>
  auto trans(Mat&& mat)
  {
    return invokeAtQP(Operation::Trans{}, std::forward<Mat>(mat));
  }

  // @}


  // binary vector operations
  // @{

  /// \brief Applies \ref Operation::Dot to two vector-valued GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto dot(Lhs&& lhs, Rhs&& rhs)
  {
    return invokeAtQP(Operation::Dot{},
      std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }

  /// \brief Applies a distance-functor to two vector-valued GridFunctions.\relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto distance(Lhs&& lhs, Rhs&& rhs)
  {
    using namespace Operation;
    return invokeAtQP(compose(TwoNorm{}, Minus{}),
      std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }

  /// \brief Applies an outer-functor to two vector-valued GridFunctions. \relates FunctorGridFunction
  template <class Lhs, class Rhs,
    REQUIRES(Concepts::AnyGridFunction<Lhs,Rhs>)>
  auto outer(Lhs&& lhs, Rhs&& rhs)
  {
    return invokeAtQP([](auto const& v, auto const& w) { return outer(v,w); },
      std::forward<Lhs>(lhs), std::forward<Rhs>(rhs));
  }

  // @}

  /** @} **/

} // end namespace AMDiS
